clear
clc
close all

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%                             Notes                                 %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Parameters only apply to image "peppers". They need to be adjusted for
%%%% other images
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Method='FD': Corresponding model in paper = IRED-SD. 
%    - Parameters: sigma1=1.5; sigma2=1; maxiter=500; tol=1e-5; K=0.0004; kappa=0.3;
% Method='FD2nd': Corresponding model in paper = A-IRED-SD
%    - Parameters: sigma1=1.5; sigma2=1; maxiter=500; tol=1e-5; K=0.0004; kappa=0.3; C=0.2;
% Method='WFT' Shrinktype='sl2': Corresponding model in paper = AMT-Haar (frame=0)/AMT-Linear (frame=1) 
%                     Accelarated counterpart is with APG=1; 
%    - Parameters (Haar):   sigma1=1.5; sigma2=0.9; maxiter=500; tol=1e-5; K=0.0005; kappa=4.8;
%                           lambda=0.01; frame=0; Level=3; 
%    - Parameters (Linear): sigma1=1.5; sigma2=0.4; maxiter=500; tol=1e-5; K=0.0005; kappa=7;
%                           lambda=0.01; frame=1; Level=3;
% Method='WFT' Shrinktype='vas': Corresponding model in paper = AST-Haar (frame=0)/AST-Linear (frame=1)
%                     Accelarated counterpart is with APG=1; 
%    - Parameters (Haar):   sigma1=1.5; sigma2=3.0; maxiter=500; tol=1e-5; K=0.001; kappa=6;
%                           lambda=0.01; frame=0; Level=3; 
%    - Parameters (Linear): sigma1=1.5; sigma2=1.0; maxiter=500; tol=1e-5; K=0.001; kappa=5;
%                           lambda=0.01; frame=1; Level=3;
% Method='WFT' Shrinktype='vs': Corresponding model in paper = IST-Haar (frame=0)/IST-Linear (frame=1) 
%                     Accelarated counterpart is with APG=1; 
%    - Parameters (Haar):   maxiter=500; tol=1e-5; K=0.0004; kappa=4.0; lambda=0.01; frame=0; Level=3; 
%    - Parameters (Linear): maxiter=500; tol=1e-5; K=0.0005; kappa=7.0; lambda=0.01; frame=1; Level=3;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


fname='./Images/peppers256.png';
addpath('./2D');
rand('state', 3000);
randn('state',3000);

Img=im2double(imread(fname));

%%%% Adjustable Parameters %%%%
sigma1=1.5;
sigma2=1.0;
maxiter=500;tol=1e-5;
K=0.001;
kappa=5.0;
% Only apply to WFT methods
lambda=0.01;
APG=1;
frame=1;Level=3; 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
ker=fspecial('gaussian',11,sigma1);
ker1=fspecial('gaussian',11,sigma2);
cker=rot90(ker,2);
blur=@(f,k)imfilter(f,k,'circular');
A = @(x) blur(x,ker);
AT = @(x) blur(x,cker);
fb=blur(Img,ker);
f=fb+max(fb(:))/20*randn(size(fb));
G = @(x,K) (1./(1+x.^2/K));
Gvas = @(x,K) (1./(1+x.^0.5/K));
hx=1;hy=1;N=size(f,1);tau=1e-4/K;mu=kappa*tau;

% Method='FD';
% Method='FD2nd';C=0.2;
% Method='WFT';shrinktype='sl2';
Method='WFT';shrinktype='vas';
% Method='WFT';shrinktype='vs';

[D,R]=GenerateFrameletFilter(frame);nBands=length(D)-1;
W  = @(x) FraDecMultiLevel2D(x,D,Level); % Frame decomposition
WT = @(x) FraRecMultiLevel2D(x,R,Level); % Frame reconstruction

tic
u=f;uout=0;psnr_curr=0;bestiter=0;ut=u;up=u;unew=u;t0=0;
for k=1:maxiter
    switch Method
        case 'FD'
            ux_m=1/hx*(u-u([end 1:end-1],:));
            ux_p=1/hx*(u([2:end 1],:)-u);
            uy_m=1/hy*(u-u(:,[end 1:end-1]));
            uy_p=1/hy*(u(:,[2:end 1])-u);
            nablau_m=sqrt(ux_m.^2+uy_m.^2);nablau_m=blur(nablau_m,ker1);
            nablau_p=sqrt(ux_p.^2+uy_p.^2);nablau_p=blur(nablau_p,ker1);
            gux_m=G(nablau_m,K);gux_p=G(nablau_p,K);
            guy_m=gux_m;guy_p=gux_p;
            uincr=1/hx*(gux_p.*ux_p-gux_m.*ux_m)+1/hy*(guy_p.*uy_p-guy_m.*uy_m);
            u=u+tau*uincr-kappa*AT(A(u)-f);
        case 'FD2nd'
            ux_m=1/hx*(u-u([end 1:end-1],:));
            ux_p=1/hx*(u([2:end 1],:)-u);
            uy_m=1/hy*(u-u(:,[end 1:end-1]));
            uy_p=1/hy*(u(:,[2:end 1])-u);
            nablau_m=sqrt(ux_m.^2+uy_m.^2);nablau_m=blur(nablau_m,ker1);
            nablau_p=sqrt(ux_p.^2+uy_p.^2);nablau_p=blur(nablau_p,ker1);
            gux_m=G(nablau_m,K);gux_p=G(nablau_p,K);
            guy_m=gux_m;guy_p=gux_p;
            uincr=1/hx*(gux_p.*ux_p-gux_m.*ux_m)+1/hy*(guy_p.*uy_p-guy_m.*uy_m);
            unew=1/(1+C)*((2+C)*u-up+(tau*uincr-kappa*AT(A(u)-f)));
            up=u;u=unew;
        case 'WFT'
            if APG==1
                t1=(1+sqrt(1+4*t0^2))/2;t0=t1;gamma=(t0-1)/t1;u=(1+gamma)*u-gamma*up;
            end
            c=W(u);
            switch shrinktype
                case 'sl2'
                    for lvl=1:Level
                        for kk=1:nBands
                            for jj=1:nBands
                                if kk==1 && jj==1
                                    Thresh{lvl}{kk,jj}=0;
                                else
                                    Thresh{lvl}{kk,jj}=4*tau*G(2*sqrt(blur(c{lvl}{kk,jj},ker1).^2+blur(c{lvl}{jj,kk},ker1).^2),K);
                                end
                            end
                        end
                    end
                case 'vs'
                    for lvl=1:Level
                        for kk=1:nBands
                            for jj=1:nBands
                                if kk==1 && jj==1
                                    Thresh{lvl}{kk,jj}=0;
                                else
                                    Thresh{lvl}{kk,jj}=lambda*0.5^(lvl-1);
                                end
                            end
                        end
                    end
                case 'vas'
                    for lvl=1:Level
                        for kk=1:nBands
                            for jj=1:nBands
                                if kk==1 && jj==1
                                    Thresh{lvl}{kk,jj}=0;
                                else
                                    Thresh{lvl}{kk,jj}=4*tau*Gvas(2*sqrt(blur(c{lvl}{kk,jj},ker1).^2+blur(c{lvl}{jj,kk},ker1).^2),K);
                                end
                            end
                        end
                    end
            end
            c=CoeffOper2D(shrinktype,c,Thresh);
            WTc=WT(c);
            if APG==1 
                unew=WTc-mu*AT(A(WTc))+mu*AT(f);up=u;u=unew;
            else
                u=WTc-mu*AT(A(WTc))+mu*AT(f);
            end
    end
    error=norm(u-ut)/norm(f);ut=u;
    figure(1);
    if mod(k,10)==0
        ax(1)=subplot(121);imshow(f,[]);
        ax(2)=subplot(122);imshow(u,[]);title(['Iter = ' num2str(k) '; PSNR = ' num2str(psnr(Img,u)) '; Error = ' num2str(error)]);drawnow;
        linkaxes(ax);
    end
    if error<tol
        break;
    end
end
if k==maxiter
    uout=u;bestiter=k;
end
ax(1)=subplot(121);imshow(f,[]);
ax(2)=subplot(122);imshow(u,[]);title(['Iter = ' num2str(k) '; PSNR = ' num2str(psnr(Img,u))]);drawnow;
linkaxes(ax);
toc